# from groq import Groq
import google.generativeai as genai
import json
from jupyter_server.base.handlers import APIHandler
from jupyter_server.utils import url_path_join
import tornado
import os
from dotenv import load_dotenv
import requests
from bs4 import BeautifulSoup
from transformers import pipeline
import sys
import traceback
from io import StringIO
from contextlib import redirect_stdout, redirect_stderr

dotenv_path = os.path.join(os.path.dirname(__file__), ".env")
load_dotenv(dotenv_path)

# Create the Groq client
# groq_api_key = os.getenv("GROQ_API_KEY")
# client = Groq(api_key=groq_api_key)

gemini_api_key = str(os.getenv("GEMINI_API_KEY"))
genai.configure(api_key=gemini_api_key)
model = genai.GenerativeModel("gemini-1.5-flash")

pipe = pipeline("summarization", model="facebook/bart-large-cnn")

# Set the system prompt
# system_prompt = {
#     "role": "system",
#     # "content": "You are a helpful assistant. You reply with short answers.",
#     "content": 'Analyze the notebook headers and the webpage text. Summarize only the webpage text and suggest a location to insert the webpage text summary into the notebook. Your response should be nothing else but a JSON-parsable string: {"indexToInsert": index_location, "title": title_or_url, "summary": summary_of_text}.',
# }

# prompt_summary = 'Analyze the notebook headers and the webpage text. Summarize only the webpage text and suggest a location to insert the webpage text summary into the notebook. Your response should be nothing else but a JSON-parsable string: "{"indexToInsert": index_location, "title": title_or_url, "summary": summary_of_text}". '
# prompt_question = 'Analyze the notebook headers and descriptions. Generate a question to ask in a lecture that is directly related to the notebook contents. Your response should be nothing else but a JSON-parsable string: "{"indexToInsert": index_location, "title": title_or_url, "summary": summary_of_text}". '
# prompt = prompt_summary

prompt_summary = 'Analyze the notebook content and the webpage text. Concisely summarize only the webpage text and suggest a location to insert the webpage text summary into the notebook. Your response should be nothing else but a JSON-parsable string: "{"indexToInsert": index_location, "title": title, "summary": summary_of_text}".'
prompt_question = '''Analyze the lecture notebook content, the code cells, variables from the executed code, and the question type. Generate a question of specified type to be asked at the specified location of the lecture that is directly related to the notebook contents before that location. For coding question, use the most recent dataset and any DataFrames related to that dataset if applicable. Then, generate an answer to the question. For coding questions, provide the answer in a code block using triple backticks with python syntax highlighting. Your response should be nothing else but a JSON-parsable string: "{"indexToInsert": notebook.activeCellIndex, "title": title, "summary": Question: \\n\\nAnswer:\\n```python\\n<code here>\\n```}". '''
# prompt = prompt_question

# chat_history = [system_prompt]

executed_results = {}


import os
import sys
import traceback
from io import StringIO


def clean_code(code):
    """
    Removes Jupyter magic commands (e.g., %%timeit) and any other invalid notebook syntax.
    """
    cleaned_lines = []
    for line in code.split("\n"):
        if not line.strip().startswith("%%"):  # Remove cell magic commands
            cleaned_lines.append(line)
    return "\n".join(cleaned_lines)


def execute_code(code, directory):
    """
    Executes Python code in the specified directory and captures both
    output (print statements) and variables defined in the execution.
    """

    # Save the original working directory
    original_directory = os.getcwd()

    # Redirect stdout and stderr to capture output
    output_buffer = StringIO()
    error_buffer = StringIO()

    # Dictionary to store the execution namespace
    local_vars = {}

    try:
        # Change to the specified directory if provided
        if directory:
            os.chdir(directory)
            if directory not in sys.path:
                sys.path.insert(0, directory)
                print("Added directory to paths")

        print("Current directory:", os.getcwd())
        print("Files in directory:", os.listdir())
        print("sys.path:", sys.path)

        print("Flushing stdout...")
        sys.stdout.flush()
        sys.stderr.flush()

        code = clean_code(code)

        # Use redirect_stdout() and redirect_stderr() to safely capture output
        with redirect_stdout(output_buffer), redirect_stderr(error_buffer):
            print("Executing code...")
            exec(code, globals())

        print("Execution completed!")

        # Extract captured output and errors
        output = output_buffer.getvalue().strip()
        print("Output:", output)
        error = error_buffer.getvalue().strip()
        print("Error:", error)

    except Exception as e:
        print(f"Exception: {e}")
        error = traceback.format_exc()
        output = ""

    finally:
        # Restore stdout, stderr, and working directory
        sys.stdout = sys.__stdout__
        sys.stderr = sys.__stderr__
        os.chdir(original_directory)

    # Remove built-in functions and modules
    executed_results = {k: v for k, v in globals().items() if not k.startswith("__")}
    print(
        "To be returned:",
        {"output": output, "error": error, "variables": executed_results},
    )

    return {"output": output, "error": error, "variables": executed_results}


class MessageHandler(APIHandler):
    @tornado.web.authenticated
    async def post(self):
        try:
            global prompt_summary
            global prompt_question
            
            # Get all variables upfront
            input_data = self.get_json_body()
            user_input = input_data.get("message", "")
            nt_content = input_data.get("notebookContent", "")
            user_notebook = str(nt_content)
            prompt_type = input_data.get("promptType", "")
            selected_cell = input_data.get("selectedCell", "")
            question_type = input_data.get("questionType", "")
            code_cells = input_data.get("notebookCodeCells", [])
            
            # Create a local prompt variable - no global modification
            current_prompt = prompt_summary if prompt_type == "summary" else prompt_question

            if prompt_type == "question":
                try:
                    # Handle code execution for questions
                    if question_type != "conceptual":
                        notebook_directory = input_data.get("notebookDirectory", "")
                        print(f"Notebook directory: {notebook_directory}")
                        
                        notebook_code_cells = input_data.get("notebookCodeCells", [])
                        print(f"Number of code cells: {len(notebook_code_cells)}")
                        
                        if notebook_code_cells:
                            execution_result = execute_code(
                                "\n\n".join(cell["content"] for cell in notebook_code_cells),
                                notebook_directory
                            )
                            print("Code execution result:", execution_result)
                            executed_results = execution_result.get("variables", {})
                        else:
                            executed_results = {}
                except Exception as e:
                    print("Error during code execution:", e)
                    traceback.print_exc()  # Print full stack trace
                    self.set_status(500)
                    self.finish(json.dumps({"error": f"Code execution failed: {str(e)}"}))
                    return

                try:
                    # Construct question prompt
                    if question_type == "conceptual":
                        current_prompt = (
                            current_prompt
                            + "\nNotebook content: "
                            + user_notebook
                            + "\nQuestion type: "
                            + question_type
                            + "\nUser question: "
                            + user_input
                        )
                    else:
                        current_prompt = (
                            current_prompt
                            + "\nNotebook content: "
                            + user_notebook
                            + "\nCode cells: "
                            + str(code_cells)
                            + "\nExecuted code variables: "
                            + str(executed_results)
                            + "\nQuestion type: "
                            + question_type
                            + "\nUser question: "
                            + user_input
                        )
                    print("Final prompt for question:", current_prompt)  # Debug log
                except Exception as e:
                    print("Error constructing question prompt:", e)
                    traceback.print_exc()
                    self.set_status(500)
                    self.finish(json.dumps({"error": f"Failed to construct prompt: {str(e)}"}))
                    return

            try:
                # Make API call
                print("\n=== LLM Request Details ===")
                print(f"Prompt Type: {prompt_type}")
                print(f"Question Type: {question_type if prompt_type == 'question' else 'N/A'}")
                print(f"Using Selected Cells: {input_data.get('useSelectedCells', False)}")
                print(f"User Input: {user_input}")
                print("\nFull Prompt being sent to LLM:")
                print("------------------------")
                print(current_prompt)
                print("------------------------\n")

                llm_response = model.generate_content(current_prompt)
                print("\n=== LLM Response ===")
                print(llm_response.text)
                print("------------------------\n")
                output = llm_response.text
                
                # Construct and send response
                response_data = {
                    "reply": output,
                    "status": "success"
                }
                print("Sending response:", response_data)
                self.finish(json.dumps(response_data))
                
            except Exception as e:
                print("Error during API call or response handling:", e)
                traceback.print_exc()
                self.set_status(500)
                self.finish(json.dumps({
                    "error": f"API or response handling failed: {str(e)}",
                    "prompt_length": len(current_prompt)
                }))
                return

        except Exception as e:
            print("Outer most layer exception:", e)
            traceback.print_exc()
            self.set_status(500)
            self.finish(json.dumps({"error": f"Request failed: {str(e)}"}))

    @tornado.web.authenticated
    def get(self):
        self.finish(json.dumps({"status": "Message service is running"}))


def setup_handlers(web_app):
    host_pattern = ".*$"

    base_url = web_app.settings["base_url"]
    route_pattern = url_path_join(base_url, "server-extension", "message")
    handlers = [
        (route_pattern, MessageHandler),
    ]
    web_app.add_handlers(host_pattern, handlers)
