import {
  JupyterFrontEnd,
  JupyterFrontEndPlugin,
  ILayoutRestorer
} from '@jupyterlab/application';

import { MainAreaWidget } from '@jupyterlab/apputils';
import { Widget } from '@lumino/widgets';
import { INotebookTracker } from '@jupyterlab/notebook';
import { requestAPI } from './handler';
import '../style/index.css';
import { NotebookPanel } from '@jupyterlab/notebook';
import { INotebookModel } from '@jupyterlab/notebook';


/**
 * Initialization data for the server-extension extension.
 */
const plugin: JupyterFrontEndPlugin<void> = {
  id: 'server-extension:plugin',
  description: 'A JupyterLab extension.',
  autoStart: true,
  requires: [INotebookTracker],
  optional: [ILayoutRestorer],
  activate: (app: JupyterFrontEnd, notebookTracker: INotebookTracker, restorer: ILayoutRestorer | null) => {
    console.log('JupyterLab extension activated :))))');

    const newWidget = () => {
      const content = new Widget();

      // Create and style the title
      const titleDiv = document.createElement('div');
      titleDiv.textContent = 'Chat Interface';
      titleDiv.classList.add('chatbot-title');
      content.node.appendChild(titleDiv);

      content.node.classList.add('chatbot-panel');

      // Create a container for the prompt and button
      const inputContainer = document.createElement('div');
      inputContainer.classList.add('input-container');

      // Add a text input box
      const inputBox = document.createElement('input');
      inputBox.type = 'url';
      inputBox.placeholder = 'Enter a valid URL...';
      inputBox.required = true;

      // Create mode toggle button
      const modeToggleButton = document.createElement('button');
      modeToggleButton.textContent = '🔗'; // URL mode icon
      modeToggleButton.classList.add('mode-toggle');
      let isUrlMode = true; // Track current mode

      // Create question type toggle button
      const questionTypeButton = document.createElement('button');
      questionTypeButton.textContent = '💭'; // Conceptual question icon
      questionTypeButton.classList.add('question-type-toggle');
      questionTypeButton.style.display = 'none'; // Hidden by default
      let isConceptualQuestion = true; // Track question type

      // Add question type toggle functionality
      questionTypeButton.addEventListener('click', () => {
        isConceptualQuestion = !isConceptualQuestion;
        if (isConceptualQuestion) {
          questionTypeButton.textContent = '💭'; // Conceptual
          questionTypeButton.title = 'Conceptual Question';
        } else {
          questionTypeButton.textContent = '💻'; // Coding
          questionTypeButton.title = 'Coding Question';
        }
      });

      // Add after questionTypeButton
      const cellScopeButton = document.createElement('button');
      cellScopeButton.textContent = '📑'; // All cells icon
      cellScopeButton.classList.add('cell-scope-toggle');
      cellScopeButton.style.display = 'none'; // Hidden by default
      cellScopeButton.title = 'Using all previous cells';
      let useAllCells = true; // Track cell scope

      // Add cell scope toggle functionality
      cellScopeButton.addEventListener('click', () => {
        useAllCells = !useAllCells;
        if (useAllCells) {
          cellScopeButton.textContent = '📑'; // All cells
          cellScopeButton.title = 'Using all previous cells';
        } else {
          cellScopeButton.textContent = '✂️'; // Selected cells
          cellScopeButton.title = 'Using selected cells only';
        }
      });

      // Update mode toggle to show/hide both buttons
      modeToggleButton.addEventListener('click', () => {
        isUrlMode = !isUrlMode;
        if (isUrlMode) {
          modeToggleButton.textContent = '🔗';
          inputBox.type = 'url';
          inputBox.placeholder = 'Enter a valid URL...';
          questionTypeButton.style.display = 'none';
          cellScopeButton.style.display = 'none';
        } else {
          modeToggleButton.textContent = '❓';
          inputBox.type = 'text';
          inputBox.placeholder = 'Enter your question...';
          questionTypeButton.style.display = 'inline';
          cellScopeButton.style.display = 'inline';
        }
      });

      // Add a send button
      const sendButton = document.createElement('button');
      sendButton.textContent = 'Submit';

      // Append input elements
      inputContainer.appendChild(modeToggleButton);
      inputContainer.appendChild(questionTypeButton); // Add question type button
      inputContainer.appendChild(cellScopeButton);
      inputContainer.appendChild(inputBox);
      inputContainer.appendChild(sendButton);
      content.node.appendChild(inputContainer);

      // Add notebook name display div after the input container
      const notebookNameDiv = document.createElement('div');
      notebookNameDiv.classList.add('notebook-name-display');
      notebookNameDiv.textContent = 'No notebook open';
      content.node.appendChild(notebookNameDiv);

      // Create chat display container
      const chatContainer = document.createElement('div');
      chatContainer.classList.add('chat-container');
      content.node.appendChild(chatContainer);

      // Create a div for Apply/Cancel buttons BELOW the chat
      const buttonContainer = document.createElement('div');
      buttonContainer.id = "chat-buttons-container"; 
      content.node.appendChild(buttonContainer);  // Append to the right panel

      let notebookDirectory: string = ''; // current notebook directory

      // Handle message sending
      const handleMessage = async () => {
        const message = inputBox.value.trim();
        if (message) {
          // Get current notebook
          const currentNotebook = notebookTracker.currentWidget;
          const notebookContent: { index: number; content: string }[] = [];

          updateNotebookDirectory();
          
          const notebookCodeCells: {
            index: number;          // Cell index in the notebook
            content: string;        // Raw code content of the cell
            isDataLoading: boolean; // Whether this cell loads a dataset
            dataframeVar: string | null; // Variable name of the loaded DataFrame (if any)
          }[] = [];
        
          const dataLoadingPattern = /(pd\.read_(csv|excel|json|html)\(|sns\.load_dataset\(|pd\.DataFrame\(|pd\.DataFrame\.from_dict\(|pd\.DataFrame\.from_records\()/; // Regular expression pattern to detect DataFrame creation
          const dataLoadingCells: number[] = []; // List to store indices of dataset-loading cells
          if (currentNotebook && currentNotebook.content) {
            const notebook = currentNotebook.content;
            const notebookModel = notebook?.model;

            if (notebookModel) {
              
              const cells = notebookModel.cells;
              const selectedCells = notebook.selectedCells; // Get selected cells

              // Handle cells differently based on mode
              if (isUrlMode) {
                // URL mode - process all cells
                const cells = notebookModel.cells;
                for (let i = 0; i < cells.length; i++) {
                    const cell = cells.get(i);
//                     if (cell.type === 'markdown' && 
//                         cell.sharedModel.getSource().includes("#")) {
                    notebookContent.push({
                        index: i + 1,
                        content: cell.sharedModel.getSource()
                    });
                }
              } else {
                // Question mode - require cell selection
                if (selectedCells.length === 0) {
                        addMessageToChat('system', '⚠️ Warning: Please select cells to generate a question about their content');
                        return;
                }
                
                for (let i = 0; i < cells.length; i++) {
                  const cell = cells.get(i);
                  const cellContent = cell.sharedModel.getSource();

                  notebookContent.push({
                      index: i + 1,
                      content: cellContent
                  });
                  
//                   // Process only selected cells for question mode
//                     selectedCells.forEach(cell => {
//                         if (cell.model.type === 'markdown' && 
//                             cell.model.sharedModel.getSource().includes("#")) {
                            
//                             const cellIndex = notebook.widgets.findIndex(widget => widget === cell);
                            
//                             notebookContent.push({
//                                 index: cellIndex + 1,
//                                 content: cell.model.sharedModel.getSource()
//                             });
//                         }
//                     });

//                     if (notebookContent.length === 0) {
//                         addMessageToChat('system', '⚠️ No markdown cells with headers (#) found in selection');
//                         return;
//                     }
                  
                  if (cell.type === 'code') {

                    const loadsData = dataLoadingPattern.test(cellContent);

                    let dataframeVar: string | null = null; // DataFrame variable name
                    if (loadsData) { 

                      dataLoadingCells.push(i + 1);

                      const dataLoadingPattern = /(\b\w+)\s*=\s*(?:pd\.read_\w+\(|sns\.load_dataset\(|pd\.DataFrame\()/;
                      const assignmentMatch = cellContent.match(dataLoadingPattern);

                      if (assignmentMatch) {
                          dataframeVar = assignmentMatch[1];  // Extract variable name
                      }
                    }
                    console.log('DataFrame detected: ' + dataframeVar);

                    notebookCodeCells.push({
                      index: i + 1,
                      content: cellContent,
                      isDataLoading: loadsData,
                      dataframeVar: dataframeVar
                    });
                  }
                }
              }

                // For URL mode, we don't need cell content <-- Ylesia: actually we do
                console.log(isUrlMode ? 'URL mode - processing all cells' : 'Selected cells content:', selectedCells);

                // Prepare the request body
//               const requestBody = {
//                 message,
//                 notebookContent,
//                 notebookCodeCells,
//                 notebookDirectory
//               };
                
                // Prepare the request body with promptType
                let relevantContent;
                if (!isUrlMode) {
                    if (useAllCells) {
                        // Get all cells up to the active cell
                        const notebook = notebookTracker.currentWidget;
                        if (notebook) {
                            const activeIndex = notebook.content.activeCellIndex;
                            const cells = [];
                            for (let i = 0; i <= activeIndex; i++) {
                                const cell = notebook.content.widgets[i];
                                if (cell) {
                                    cells.push({
                                        index: i + 1,
                                        content: cell.model.sharedModel.getSource()
                                    });
                                }
                            }
                            relevantContent = cells;
                        }
                    } else {
                        // Get only selected cells
                        relevantContent = selectedCells.map((cell, index) => ({
                            index: index + 1,
                            content: cell.model.sharedModel.getSource()
                        }));
                    }
                }
                
                const requestBody = {
                  message: message,
                  notebookContent: relevantContent || notebookContent,
                  promptType: isUrlMode ? 'summary' : 'question',
                  selectedCell: !isUrlMode ? selectedCells[selectedCells.length - 1].model.sharedModel.getSource() : null,
                  questionType: !isUrlMode ? (isConceptualQuestion ? 'conceptual' : 'coding') : null,
                  useSelectedCells: !useAllCells,
                  activeCellIndex: notebookTracker.currentWidget?.content.activeCellIndex ?? 0,
                  ...(isUrlMode ? {} : { 
                    notebookDirectory,
                    notebookCodeCells
                  })
                };
                console.log("Selected Cell Index: " + selectedCells[selectedCells.length - 1].model.sharedModel.getSource());
                console.log(requestBody);

                addMessageToChat('user', message);
                inputBox.value = '';

                try {

                  // Making POST request to message endpoint
                  const response = await requestAPI<any>('message', {
                    method: 'POST',
                    body: JSON.stringify(requestBody)
                  });

                  console.log("Working up to here");
                  // console.log(response.reply)
                  console.log(response)


                // const llmOutput = JSON.parse(response.reply)


                const croppedString = response.reply.substring(7, response.reply.length - 4);

                const llmOutput = JSON.parse(croppedString);
                const indexToInsert = notebook.activeCellIndex + 1;
                console.log("Index to insert: " + indexToInsert);
                const summary = llmOutput.summary;


                addMessageToChat('assistant', 'Location: ' + indexToInsert + '\n\nSummary: ' + summary);
                
                console.log("Working up to here #2");
                const numCells = notebookModel.sharedModel.cells.length;
                const safeIndex = Math.min(indexToInsert, numCells);  // Ensure index is within bounds

                console.log(`Inserting new cell at index ${safeIndex} with summary:`, summary);

                const pageTitle = llmOutput.title || inputBox.value;  // Use title if available, otherwise fallback to URL

                notebookModel.sharedModel.insertCell(safeIndex, {
                    cell_type: 'markdown',
                    source: `### ${pageTitle}\n\n${summary}`,
                    metadata: { temporary: true }  // Mark as temporary
                });

                attachButtonsBelowChat(safeIndex, notebookModel);


                // notebookModel.sharedModel.insertCell(safeIndex, {
                //     cell_type: 'markdown',
                //     source: `### ${pageTitle}\n\n${summary}`,
                //     metadata: { temporary: true }  // Mark as temporary
                // });


                // Attach listeners for buttons
                setTimeout(() => attachButtonListeners(safeIndex, notebookModel), 100);
              
              } catch (error) {
                console.error('Failed to get response:', error);
                addMessageToChat('system', 'Error: Failed to get response');
              }

              // helper functions
              function attachButtonListeners(index: number, notebookModel: INotebookModel) {
                // Get all cells
                const notebookWidget = document.querySelector(".jp-Notebook");
                if (!notebookWidget) return;
            
                // Find the cell at the given index
                const cellNodes = notebookWidget.querySelectorAll(".jp-Cell");
                if (index >= cellNodes.length) return;
                const cellNode = cellNodes[index];
            
                // Find buttons inside the cell
                const applyBtn = cellNode.querySelector(".apply-btn") as HTMLButtonElement;
                const cancelBtn = cellNode.querySelector(".cancel-btn") as HTMLButtonElement;
            
                if (applyBtn) {
                    applyBtn.addEventListener("click", () => applyChanges(index, notebookModel));
                }
                if (cancelBtn) {
                    cancelBtn.addEventListener("click", () => cancelChanges(index, notebookModel));
                }
              }
              function applyChanges(index: number, notebookModel: INotebookModel) {
                console.log(`Applying changes for cell at index ${index}`);
            
                // Find the cell
                const cell = notebookModel.sharedModel.cells[index];
                if (!cell) return;
            
                // Remove temporary metadata
                delete cell.metadata.temporary;
            
                // Remove the buttons from the cell content
                (cell as any).setSource(
                    (cell as any).getSource()
                        .replace(/\*\*🔹 Suggested Summary:\*\*\n\n/, "")  // Remove label
                        .replace(/\n\n<button class="apply-btn">✅ Apply<\/button> <button class="cancel-btn">❌ Cancel<\/button>/, "")
                );
            
                //Remove the buttons from the chat area
                removeChatButtons();
            
                console.log(`Cell at index ${index} is now permanent.`);
              }
            
              function cancelChanges(index: number, notebookModel: INotebookModel) {
                console.log(`Cancelling changes and deleting cell at index ${index}`);
                
                // Remove the inserted summary cell
                notebookModel.sharedModel.deleteCell(index);
            
                // Remove the buttons from the chat area
                removeChatButtons();
            
                console.log(`Deleted cell at index ${index}`);
              }

              function removeChatButtons() {
                const panel = document.getElementById("chat-buttons-container");
                if (panel) {
                    panel.innerHTML = "";  // Clear the buttons from the UI
                }
              }  
          
              function attachButtonsBelowChat(index: number, notebookModel: INotebookModel) {
                console.log(`Adding buttons below chat for cell at index ${index}`);
            
                // Ensure the chat buttons container exists
                let panel = document.getElementById("chat-buttons-container");
                if (!panel) {
                    console.error("Chat buttons container not found! Creating one...");
                    panel = document.createElement("div");
                    panel.id = "chat-buttons-container";
                    panel.classList.add("suggested-summary-buttons"); // Ensure correct CSS class
                    // panel.innerHTML = "<p><b>Summary Actions</b></p>";
                    document.body.appendChild(panel);  // Attach to body as a fallback
                }
            
                // Create Apply button
                const applyBtn = document.createElement("button");
                applyBtn.textContent = "✅ Apply";
                applyBtn.className = "apply-btn"; // Assign correct class
                applyBtn.onclick = () => applyChanges(index, notebookModel);
            
                // Create Cancel button
                const cancelBtn = document.createElement("button");
                cancelBtn.textContent = "❌ Cancel";
                cancelBtn.className = "cancel-btn"; // Assign correct class
                cancelBtn.onclick = () => cancelChanges(index, notebookModel);
            
                // Add buttons to the panel
                panel.appendChild(applyBtn);
                panel.appendChild(cancelBtn);
            
                console.log(`Buttons added below chat.`);
              }
            }
          }
        }
      };


      // Add event listeners
      // sendButton.addEventListener('click', handleMessage);
      sendButton.addEventListener('click', () => {
        if (isUrlMode) {
          // URL mode validation
          if (inputBox.validity.valid) {
            console.log('Valid URL:', inputBox.value);
            handleMessage();
          } else {
            addMessageToChat('system', 'Error: Please input a valid link to get response');
          }
        } else {
          // Question mode - no URL validation needed
          if (inputBox.value.trim()) {
            handleMessage();
          } else {
            addMessageToChat('system', 'Error: Please enter a question');
          }
        }
      });
      inputBox.addEventListener('keypress', (event) => {
        if (event.key === 'Enter') {
          if (isUrlMode) {
            // URL mode validation
            if (inputBox.validity.valid) {
              console.log('Valid URL:', inputBox.value);
              handleMessage();
            } else {
              addMessageToChat('system', 'Error: Please input a valid link to get response');
            }
          } else {
            // Question mode - no URL validation needed
            if (inputBox.value.trim()) {
              handleMessage();
            } else {
              addMessageToChat('system', 'Error: Please enter a question');
            }
          }
        }
      });
    
      // Function to add messages to chat
      const addMessageToChat = (role: string, text: string) => {
        const messageDiv = document.createElement('div');
        messageDiv.classList.add('chat-message', role);

        // messageDiv.textContent = text;
        messageDiv.innerHTML = text.replace(/\n/g, "<br>"); 

        chatContainer.appendChild(messageDiv);
        chatContainer.scrollTop = chatContainer.scrollHeight;
      };

      // Update notebook name
      const updateNotebookName = () => {
        // Get current notebook
        const currentNotebook = notebookTracker.currentWidget;

        if (currentNotebook instanceof NotebookPanel) {
          const name = currentNotebook.title.label;
          // Get notebook name
          notebookNameDiv.textContent = `Notebook: ${name}`;
        } else {
          notebookNameDiv.textContent = 'No notebook detected...';
        }
      };

      const updateNotebookDirectory = () => {
        const currentNotebook = notebookTracker.currentWidget;
    
        if (currentNotebook && currentNotebook.context) {
            const notebookPath = currentNotebook.context.path;
            // console.log("notebookPath: " + notebookPath);
            // notebookDirectory = notebookPath.substring(0, notebookPath.lastIndexOf('/'));
            
            const lastSlashIndex = notebookPath.lastIndexOf('/');
            // If '/' is found, extract the directory path; otherwise, default to "."
            notebookDirectory = lastSlashIndex !== -1 
                ? notebookPath.substring(0, lastSlashIndex) 
                : ".";  // Current directory if no `/` is found

            console.log("Notebook Directory updated:", notebookDirectory);
        }
      };
    
    
      // Listen for changes in the active notebook
      notebookTracker.currentChanged.connect(() => {
        updateNotebookName();
        updateNotebookDirectory;
      });

      // Initial update
      updateNotebookName();
      updateNotebookDirectory;

      const widget = new MainAreaWidget({ content });
      widget.id = 'chatbot-widget';
      widget.title.label = 'Chat';
      widget.title.closable = true;

      // Add widget to the right panel
      app.shell.add(widget, 'right');

      if (restorer) {
        restorer.add(widget, 'chatbot-widget');
      }

      // Add this CSS for highlighting new cells
      const style = document.createElement('style');
      style.textContent = `
        .jp-Cell.highlighted {
          background-color: rgba(255, 255, 0, 0.1) !important;  /* Light yellow highlight */
          border-left: 3px solid #ffd700 !important;  /* Gold border */
          margin-left: -3px !important;
          padding-left: 12px !important;
          transition: background-color 0.3s ease;
        }
      `;
      document.head.appendChild(style);

      return widget;
    };

    // Create and display the widget
    newWidget();
  }
};

export default plugin;
